import re
from typing import Any

from custom_python_logger import get_logger
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from rest_framework.exceptions import ValidationError
from rest_framework.request import Request

from drf_easy_crud.const import INTEGER_FIELDS, OPERATOR_MAPPING, RELATED_LOOKUP, RESERVED_PARAMS

logger = get_logger(__name__)


class FilterUtils:
    """Utility class for applying filtering and pagination to Django REST Framework views."""

    @staticmethod
    def _wildcard_to_regex(pattern: str) -> str:
        """Convert wildcard pattern to regex pattern."""

        return re.escape(pattern).replace(r"\*", ".*")

    @staticmethod
    def _has_middle_wildcard(value: str) -> bool:
        """Check if wildcard appears in the middle of the string."""

        return len(value) > 2 and "*" in value[1:-1]

    @staticmethod
    def _apply_text_wildcard_filter(
        queryset: models.QuerySet,
        param_name: str,
        param_value: str,
    ) -> models.QuerySet:
        """Apply wildcard filtering to text fields."""

        if "*" not in param_value:
            return queryset.filter(**{f"{param_name}__iexact": param_value})
        if FilterUtils._has_middle_wildcard(param_value) or param_value.count("*") > 1:
            return queryset.filter(**{f"{param_name}__iregex": FilterUtils._wildcard_to_regex(param_value)})
        if param_value.startswith("*") and param_value.endswith("*"):
            if clean_value := param_value.strip("*"):
                return queryset.filter(**{f"{param_name}__icontains": clean_value})
        if param_value.startswith("*"):
            if clean_value := param_value.lstrip("*"):
                return queryset.filter(**{f"{param_name}__iendswith": clean_value})
        if param_value.endswith("*"):
            if clean_value := param_value.rstrip("*"):
                return queryset.filter(**{f"{param_name}__istartswith": clean_value})
        return queryset

    @staticmethod
    def _convert_number_value(value: str, field: models.Field) -> int | float:
        """Convert string to int or float based on field type."""

        return int(value) if isinstance(field, INTEGER_FIELDS) else float(value)

    @staticmethod
    def _apply_number_filter(
        queryset: models.QuerySet,
        param_name: str,
        param_value: str,
        field: models.Field,
    ) -> models.QuerySet:
        """Apply filtering to number fields with support for comparison operators."""

        param_value = param_value.strip()
        lookup_suffix = ""
        value_str = param_value
        for operator, suffix in OPERATOR_MAPPING.items():
            if param_value.startswith(operator):
                lookup_suffix = suffix
                value_str = param_value[len(operator) :].strip()
                break

        try:
            value = FilterUtils._convert_number_value(value_str, field)
            filter_key = f"{param_name}{lookup_suffix}" if lookup_suffix else param_name
            return queryset.filter(**{filter_key: value})
        except (ValueError, TypeError):
            return queryset.none()

    @staticmethod
    def _apply_wildcard_to_field(
        queryset: models.QuerySet,
        param_name: str,
        param_value: str,
        field: models.Field,
        django_models: Any,
    ) -> models.QuerySet:
        """Apply wildcard filtering to a specific field.

        Supports:
        - Text fields: wildcard patterns (*test*, test*, *test, t*t)
        - Number fields: comparison operators (>=10, <=100, >5, <20) and ranges (10-20)
        - Other fields: exact match

        Args:
            queryset: The queryset to filter.
            param_name: The parameter name (field or lookup path).
            param_value: The parameter value (may contain wildcards or operators).
            field: The Django field object.
            django_models: Django models module.

        Returns:
            Filtered queryset.
        """
        is_number_field = isinstance(
            field,
            django_models.IntegerField
            | django_models.BigIntegerField
            | django_models.SmallIntegerField
            | django_models.PositiveIntegerField
            | django_models.PositiveSmallIntegerField
            | django_models.FloatField
            | django_models.DecimalField,
        )

        if is_number_field:
            return FilterUtils._apply_number_filter(queryset, param_name, param_value, field)

        if isinstance(field, django_models.CharField | django_models.TextField):  # is_text_field
            return FilterUtils._apply_text_wildcard_filter(queryset, param_name, param_value)

        return queryset.filter(**{param_name: param_value})  # For other field types (bool, date, etc.), use exact match

    @staticmethod
    def _get_lookup_target_field(model: type[models.Model], lookup_path: str) -> models.Field | None:
        """Get the target field from a Django lookup path."""

        parts = lookup_path.split(RELATED_LOOKUP)
        current_model = model

        for part in parts[:-1]:
            try:
                field = current_model._meta.get_field(part)  # pylint: disable=W0212
                if isinstance(field, (models.ForeignKey | models.OneToOneField)):
                    current_model = field.related_model
                elif isinstance(field, models.ManyToManyField):
                    current_model = field.related_model
                else:
                    return None
            except FieldDoesNotExist:
                return None

        # Get the final field
        try:
            return current_model._meta.get_field(parts[-1])  # pylint: disable=W0212
        except FieldDoesNotExist:
            return None

    @staticmethod
    def apply_wildcard_filtering(
        queryset: models.QuerySet,
        request: Request,
    ) -> models.QuerySet:
        """Apply generic wildcard filtering directly to queryset."""

        query_params = request.query_params.copy()
        model = queryset.model
        model_fields = {f.name for f in model._meta.get_fields()}  # pylint: disable=W0212

        filtered_queryset = queryset
        for param_name, param_value in query_params.items():
            if param_name in RESERVED_PARAMS or not param_value:  # Skip reserved params and empty values
                continue

            if RELATED_LOOKUP in param_name:  # Check if this is a ForeignKey lookup (contains "__")
                if not (target_field := FilterUtils._get_lookup_target_field(model=model, lookup_path=param_name)):
                    raise ValidationError({param_name: f"Field '{param_name}' does not exist"})
            else:  # Direct field lookup
                if param_name not in model_fields:
                    raise ValidationError({param_name: f"Field '{param_name}' does not exist on {model.__name__}"})

                # Get field type to determine appropriate lookup
                try:
                    target_field = model._meta.get_field(param_name)  # pylint: disable=W0212
                except FieldDoesNotExist as err:
                    raise ValidationError(
                        {param_name: f"Field '{param_name}' does not exist on {model.__name__}"}
                    ) from err

            filtered_queryset = FilterUtils._apply_wildcard_to_field(
                queryset=filtered_queryset,
                param_name=param_name,
                param_value=param_value,
                field=target_field,
                django_models=models,
            )
        return filtered_queryset



import os
from pathlib import Path

IGNORED_DIRS = [".venv"]
IGNORED_FILES = ["__init__.py"]


def clean_migrations_folder(root_dir: Path) -> None:
    for dirpath, dirnames, _ in os.walk(root_dir):
        if any(ignored_dir in dirpath for ignored_dir in IGNORED_DIRS):
            continue

        if "migrations" in dirnames:
            migrations_dir = os.path.join(dirpath, "migrations")
            print(f"cleaning: {migrations_dir}")

            for filename in os.listdir(migrations_dir):
                file_path = os.path.join(migrations_dir, filename)
                if filename in IGNORED_FILES and os.path.isfile(file_path):
                    continue
                try:
                    os.remove(file_path)
                    print(f"delete: {file_path}")
                except Exception as e:
                    print(f"cant find {file_path}: {e}")


def main() -> None:
    root_dir = Path(__file__).parent.parent.parent
    print(f"Root directory: {root_dir}")
    clean_migrations_folder(root_dir=root_dir)


if __name__ == "__main__":
    main()



from custom_python_logger import build_logger
from python_github_plus import GitHubClient

logger = build_logger(__name__)


def run_pytest_marker_workflow(
    branch: str | None,
    marker: str,
    repo_full_name: str = "aviz92/dev-template-repository",
    workflow_name: str = "Pytest by Marker",
) -> None:
    github_client = GitHubClient(repo_full_name=repo_full_name)

    logger.debug("Running Pytest workflow with:" f"Marker: {marker}" f"Branch: {branch}" f"Repo: {repo_full_name}")
    _ = github_client.workflow.trigger(workflow_name=workflow_name, branch_name=branch, inputs={"marker": marker})
    logger.info("Workflow triggered successfully!")


def main() -> None:
    logger.info("This script triggers the Pytest workflow on GitHub with a specified marker.")

    # parser = argparse.ArgumentParser(description="Trigger Pytest workflow on GitHub with a marker.")
    # parser.add_argument(
    #     "branch",
    #     nargs="?",
    #     help="Branch to run workflow on (defaults to current branch)",
    #     # default="branch-name",
    # )
    # parser.add_argument(
    #     "marker",
    #     nargs="?",
    #     help="The marker to run (e.g., unit1)",
    #     # default="marker-name",
    # )
    # branch = parser.parse_args().branch
    # marker = parser.parse_args().marker
    #
    # run_pytest_marker_workflow(
    #     branch=branch,
    #     marker=marker,
    # )


if __name__ == "__main__":
    main()



from dotenv import load_dotenv

from drf_easy_crud.crud_utils import CRUDUtils, StandardResultsSetPagination
from drf_easy_crud.filter_utils import FilterUtils

load_dotenv()

__all__ = [
    "StandardResultsSetPagination",
    "CRUDUtils",
    "FilterUtils",
]



import requests
from bs4 import BeautifulSoup

url = "https://pypi.org/user/aviz/"
PYPI_JSON = "https://pypi.org/pypi/{}/json"


def get_latest_version(package: str) -> str:
    session = requests.Session()
    session.headers.update({"User-Agent": "uv-upgrade-script"})
    resp = session.get(PYPI_JSON.format(package))
    resp.raise_for_status()
    return resp.json()["info"]["version"]


def get_all_pypi_projects_latest_versions() -> dict:
    response = requests.get(url)
    soup = BeautifulSoup(response.text, "html.parser")
    projects_res = soup.find_all("h3", class_="package-snippet__title")

    projects = {}
    for _, project in enumerate(projects_res, 1):
        if "Archived" not in project.text and "python-test-aviz" not in project.text:
            _project_name = project.text.strip()
            projects[_project_name] = get_latest_version(_project_name)
    return projects


def build_uv_command(projects: dict) -> str:
    command = "uv add "
    for project, latest_version in projects.items():
        command += f"{project}=={latest_version} "
    return command.strip()


def main() -> None:
    projects = get_all_pypi_projects_latest_versions()
    for project, latest_version in projects.items():
        print(f"{project}=={latest_version}")
        # print(f'"{project}=={latest_version}",')

    uv_cmd = build_uv_command(projects)
    print(uv_cmd)
    print()


if __name__ == "__main__":
    main()



from django.db import models

RESERVED_PARAMS = {"page", "page_size", "ordering", "format"}

RELATED_LOOKUP = "__"

OPERATOR_MAPPING = {  # pylint: disable=R6101
    ">=": ("__gte", 2),
    "<=": ("__lte", 2),
    ">": ("__gt", 1),
    "<": ("__lt", 1),
}

DEFAULT_PAGE_SIZE = 20
DEFAULT_MAX_PAGE_SIZE = 100
PAGE_SIZE_QUERY_PARAM = "page_size"

INTEGER_FIELDS = (
    models.IntegerField,
    models.BigIntegerField,
    models.SmallIntegerField,
    models.PositiveIntegerField,
    models.PositiveSmallIntegerField,
)



import os
import shutil
from pathlib import Path

IGNORED_DIRS = [".venv"]


def delete_pycache_folder(root_dir: Path) -> None:
    for dirpath, dirnames, _ in os.walk(root_dir):
        if any(ignored_dir in dirpath for ignored_dir in IGNORED_DIRS):
            continue

        if "__pycache__" in dirnames:
            pycache_dir = os.path.join(dirpath, "__pycache__")
            print(f"cleaning: {pycache_dir}")
            try:
                shutil.rmtree(pycache_dir)
                print(f"delete: {pycache_dir}")
            except Exception as e:
                print(f"cant find {pycache_dir}: {e}")


def main() -> None:
    root_dir = Path(__file__).parent.parent.parent
    print(f"Root directory: {root_dir}")
    delete_pycache_folder(root_dir)
    delete_pycache_folder(root_dir=root_dir)


if __name__ == "__main__":
    main()



from collections.abc import Callable
from typing import Any

from custom_python_logger import get_logger
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import IntegrityError, models
from rest_framework import status
from rest_framework.pagination import PageNumberPagination
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer

from drf_easy_crud.const import DEFAULT_MAX_PAGE_SIZE, DEFAULT_PAGE_SIZE, PAGE_SIZE_QUERY_PARAM
from drf_easy_crud.filter_utils import FilterUtils

logger = get_logger(__name__)


class StandardResultsSetPagination(PageNumberPagination):
    """Standard pagination class for list endpoints."""

    page_size = DEFAULT_PAGE_SIZE
    page_size_query_param = PAGE_SIZE_QUERY_PARAM
    max_page_size = DEFAULT_MAX_PAGE_SIZE


class CRUDUtils:
    """Enterprise-grade utility class providing comprehensive CRUD operations for Django REST Framework views."""

    @staticmethod
    def _get_instance_by_pk(
        model_class: type[models.Model],
        pk: Any,
    ) -> models.Model | None:
        """Get a model instance by primary key with optional performance optimizations."""

        try:
            queryset = model_class.objects.all()
            return queryset.get(pk=pk)
        except model_class.DoesNotExist:
            logger.warning(f"{model_class.__name__} with pk={pk} not found")
            return None

    @staticmethod
    def apply_pagination(
        queryset: models.QuerySet,
        request: Request,
        serializer_class: type[ModelSerializer],
        pagination_class: type[PageNumberPagination] | None = None,
    ) -> Response:
        """Apply pagination to a queryset and return a paginated response."""

        pagination_class = pagination_class or StandardResultsSetPagination

        paginator = pagination_class()
        if (page := paginator.paginate_queryset(queryset=queryset, request=request)) is not None:
            serializer = serializer_class(page, context={"request": request}, many=True)
            return paginator.get_paginated_response(serializer.data)

        # No pagination, return all results
        serializer = serializer_class(queryset, context={"request": request}, many=True)
        return Response(serializer.data)

    @staticmethod
    def _apply_filtering(
        queryset: models.QuerySet,
        request: Request,
    ) -> models.QuerySet:
        """Apply filtering to queryset using django-filter or generic wildcard filtering."""

        if request.query_params:
            return FilterUtils.apply_wildcard_filtering(queryset=queryset, request=request)
        return queryset

    @staticmethod
    def _build_queryset(
        model_class: type[models.Model],
        queryset_hook: Callable | None = None,
    ) -> models.QuerySet:  # constant filter: queryset_hook=lambda: FirstApp.objects.filter(is_active=True)
        """Build the initial queryset for list retrieval, allowing for optional customization via a hook."""

        return queryset_hook() if queryset_hook else model_class.objects.all()

    @staticmethod
    def get(
        request: Request,
        model_class: type[models.Model],
        serializer_class: type[ModelSerializer],
        queryset_hook: Callable | None = None,
        ordering_field: str | None = "pk",
        pagination_class: type[PageNumberPagination] | None = None,
        **kwargs: Any,
    ) -> Response:
        """Retrieve a single instance or paginated list of instances."""

        if pk := kwargs.get("pk"):
            if not (instance := CRUDUtils._get_instance_by_pk(model_class=model_class, pk=pk)):
                return Response(status=status.HTTP_404_NOT_FOUND)
            serializer = serializer_class(instance, context={"request": request})
            return Response(serializer.data)
        queryset = CRUDUtils._build_queryset(model_class=model_class, queryset_hook=queryset_hook)
        queryset = CRUDUtils._apply_filtering(queryset=queryset, request=request)
        queryset = queryset.order_by(ordering_field)
        return CRUDUtils.apply_pagination(
            queryset=queryset,
            request=request,
            serializer_class=serializer_class,
            pagination_class=pagination_class,
        )

    @staticmethod
    def post(
        request: Request,
        serializer_class: type[ModelSerializer],
        **kwargs: Any,
    ) -> Response:
        """Create a new instance."""

        serializer = serializer_class(data=request.data, context={"request": request})
        if serializer.is_valid():
            try:
                serializer.save(**kwargs)
                model_name = serializer_class.Meta.model.__name__
                instance_id = serializer.data.get("id")
                logger.info(f"Created {model_name} with id={instance_id}")
                return Response(serializer.data, status=status.HTTP_201_CREATED)
            except (IntegrityError, DjangoValidationError) as e:
                logger.error(f"Database error creating {serializer_class.Meta.model.__name__}: {str(e)}")
                return Response(
                    {"error": "Database constraint violation", "detail": str(e)}, status=status.HTTP_400_BAD_REQUEST
                )
        model_name = serializer_class.Meta.model.__name__
        logger.warning(f"Validation failed for {model_name}: {serializer.errors}")
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @staticmethod
    def _update_instance(
        request: Request,
        model_class: type[models.Model],
        serializer_class: type[ModelSerializer],
        pk: Any,
        partial: bool,
        **kwargs: Any,
    ) -> Response:
        """Internal method to update an instance."""

        instance = CRUDUtils._get_instance_by_pk(model_class=model_class,pk=pk)
        if not instance:
            return Response(status=status.HTTP_404_NOT_FOUND)

        serializer = serializer_class(instance, data=request.data, context={"request": request}, partial=partial)
        if serializer.is_valid():
            try:
                serializer.save(**kwargs)
                update_type = "Partially updated" if partial else "Updated"
                logger.info(f"{update_type} {model_class.__name__} with pk={pk}")
                return Response(serializer.data)
            except (IntegrityError, DjangoValidationError) as e:
                logger.error(f"Database error updating {model_class.__name__} pk={pk}: {str(e)}")
                return Response(
                    {"error": "Database constraint violation", "detail": str(e)}, status=status.HTTP_400_BAD_REQUEST
                )
        logger.warning(f"Validation failed for {model_class.__name__} pk={pk}: {serializer.errors}")
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @staticmethod
    def put(
        request: Request,
        model_class: type[models.Model],
        serializer_class: type[ModelSerializer],
        **kwargs: Any,
    ) -> Response:
        """Update an existing instance with full replacement (PUT semantics)."""

        if not (pk := kwargs.pop("pk", None)):
            logger.warning(f"PUT request missing pk for {model_class.__name__}")
            return Response(status=status.HTTP_404_NOT_FOUND)
        return CRUDUtils._update_instance(
            request=request,
            model_class=model_class,
            serializer_class=serializer_class,
            pk=pk,
            partial=False,
            **kwargs,
        )

    @staticmethod
    def patch(
        request: Request,
        model_class: type[models.Model],
        serializer_class: type[ModelSerializer],
        **kwargs: Any,
    ) -> Response:
        """Partially update an existing instance (PATCH semantics)."""

        if not (pk := kwargs.pop("pk", None)):
            logger.warning(f"PATCH request missing pk for {model_class.__name__}")
            return Response(status=status.HTTP_404_NOT_FOUND)
        return CRUDUtils._update_instance(
            request=request,
            model_class=model_class,
            serializer_class=serializer_class,
            pk=pk,
            partial=True,
            **kwargs,
        )

    @staticmethod
    def delete(
        model_class: type[models.Model],
        **kwargs: Any,
    ) -> Response:
        """Delete an existing instance."""

        if not (pk := kwargs.get("pk")):
            logger.warning(f"DELETE request missing pk for {model_class.__name__}")
            return Response(status=status.HTTP_404_NOT_FOUND)

        if not (instance := CRUDUtils._get_instance_by_pk(model_class, pk)):
            return Response(status=status.HTTP_404_NOT_FOUND)

        try:
            instance.delete()
            logger.info(f"Deleted {model_class.__name__} with pk={pk}")
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception as e:
            logger.error(f"Error deleting {model_class.__name__} pk={pk}: {str(e)}")
            return Response(
                {"error": "Failed to delete instance", "detail": str(e)}, status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )
